Require Import ZArith List String Ascii extracted_io syntax.

Inductive lex_state : Type :=
  Start | In_num | In_id | In_spec.

Inductive token : Type :=
  T_id (_ : string) | T_num (_ : Z) | T_variables | T_in | T_end | T_comma
| T_semicolumn | T_assign | T_lt | T_plus |  T_minus | T_while | T_mult
| T_do | T_done | T_open | T_close | T_open_B | T_close_B | T_open_S
| T_close_S | T_conj | T_bang.

Open Scope Z_scope.
Open Scope char_scope.
Open Scope string_scope.

Inductive char_category : Type :=
 Digit | Alpha | Space | Other.

Fixpoint Z_of_rev_string (s:string) : Z :=
 match s with
   "" => 0
 | String a tl => digit_value a+10*Z_of_rev_string tl
 end.

Definition T_num_of_rev_string (s:string) : token :=
  T_num (Z_of_rev_string s).

Fixpoint rev_string_aux (s acc:string) {struct s}: string :=
 match s with "" => acc | String a s' => rev_string_aux s' (String a acc) end.

Definition rev_string (s:string) := rev_string_aux s "".

Definition string_token (rev_buf :string) : token :=
  let buf := rev_string rev_buf in
  if string_dec buf "variables" then
     T_variables
  else if string_dec buf "in" then
     T_in
  else if string_dec buf "end" then
     T_end
  else if string_dec buf "while" then
     T_while
  else if string_dec buf "do" then
     T_do
  else if string_dec buf "done" then
     T_done
  else T_id buf.

Fixpoint special_tokens (rev_buf:string) (acc:list token) {struct rev_buf}: 
  option (list token) :=
  match rev_buf with
    "" => Some acc
  | String ";" tl => special_tokens tl (T_semicolumn::acc)
  | String "!" tl => special_tokens tl (T_bang::acc)
  | String "," tl => special_tokens tl (T_comma::acc)
  | String "=" (String ":" tl) => special_tokens tl (T_assign::acc)
  | String "<" tl => special_tokens tl (T_lt::acc)
  | String "+" tl => special_tokens tl (T_plus::acc)
  | String "*" tl => special_tokens tl (T_mult::acc)
  | String "(" tl => special_tokens tl (T_open::acc)
  | String ")" tl => special_tokens tl (T_close::acc)
  | String "{" tl => special_tokens tl (T_open_B::acc)
  | String "}" tl => special_tokens tl (T_close_B::acc)
  | String "[" tl => special_tokens tl (T_open_S::acc)
  | String "]" tl => special_tokens tl (T_close_S::acc)
  | String "-" tl => special_tokens tl (T_minus::acc)
  | String "\" (String "/" tl) => special_tokens tl (T_conj::acc)
  | _ => None
  end.

Definition ascii_code (a : Ascii.ascii) : Z :=
  let f := (fun (b:bool) x => 2 * x + if b then 1 else 0) in
    let (b1,b2,b3,b4,b5,b6,b7,b8) := a in
      f b1 (f b2 (f b3 (f b4 (f b5 (f b6 (f b7 (f b8 0))))))).

Definition A_code := Eval vm_compute in ascii_code "A"%char.
Definition Z_code := Eval vm_compute in ascii_code "Z"%char.
Definition a_code := Eval vm_compute in ascii_code "a"%char.
Definition z_code := Eval vm_compute in ascii_code "z"%char.
Definition under_code := Eval vm_compute in ascii_code "_"%char.

Definition is_alpha (a:ascii) : bool :=
  match ascii_code a ?= A_code with
    Lt => false
  | Eq => true
  | Gt => match ascii_code a ?= Z_code with
      Gt => match ascii_code a ?= under_code with
        Eq => true
      | _ => match ascii_code a ?= a_code with
          Lt => false
        | Eq => true
        | Gt => match ascii_code a ?= z_code with
            Gt => false
          | _ => true
          end
        end
      end
    | _ => true
    end
  end.

Definition is_space (a:ascii) : bool :=
  if ascii_dec a " " then true
  else if ascii_dec a "010" then true
  else if ascii_dec a "013" then true
  else if ascii_dec a "009" then true
  else false.

Definition cat_of_char (a:ascii) : char_category :=
  if is_digit a then
     Digit
  else if is_alpha a then
     Alpha
  else if is_space a then
     Space
  else Other.

Fixpoint tokenize (s:string)(st:lex_state)(rev_buf:string) {struct s}
    :list token :=
  match s with
    "" => 
      match st with
        Start => nil
      | In_num => T_num_of_rev_string rev_buf::nil
      | In_id => string_token rev_buf::nil
      | In_spec => match special_tokens rev_buf nil with
                     Some ts => ts | None => nil
                   end
      end
  | String a s' =>
      match st with
        Start => 
          match cat_of_char a with
            Digit =>
            tokenize s' In_num (String a "")
          | Alpha =>
             tokenize s' In_id (String a "")
          | Space =>
             tokenize s' Start rev_buf
          | Other =>
            tokenize s' In_spec (String a "")
          end
      | In_num =>
          match cat_of_char a with
            Digit =>
            tokenize s' In_num (String a rev_buf)
          | Alpha =>
            T_num (Z_of_rev_string rev_buf)::
            tokenize s' In_id (String a "")
          | Space =>
            T_num (Z_of_rev_string rev_buf)::tokenize s' Start ""
          | Other =>
            T_num (Z_of_rev_string rev_buf)::tokenize s' In_spec (String a "")
          end
      | In_id => 
          match cat_of_char a with
            Digit =>
            tokenize s' In_id (String a rev_buf)
          | Alpha =>
             tokenize s' In_id (String a rev_buf)
          | Space =>
             string_token rev_buf::tokenize s' Start rev_buf
          | Other =>
             string_token rev_buf::tokenize s' In_spec (String a "")
          end
      | In_spec =>
          match cat_of_char a with
            Digit =>
            match special_tokens rev_buf nil with
              Some ts => (ts++tokenize s' In_num (String a ""))%list
            | None => nil
            end
          | Alpha =>
            match special_tokens rev_buf nil with
              Some ts => (ts++tokenize s' In_id (String a ""))%list
            | None => nil
            end
          | Space =>
            match special_tokens rev_buf nil with
              Some ts => (ts++tokenize s' Start "")%list
            | None => nil
            end
          | Other =>
            tokenize s' In_spec (String a rev_buf)
          end
      end
  end.

Fixpoint parse_variables (l:list token) :
  option ((list (string*Z))*list token) :=
  match l with
    T_id s::T_minus::T_num n::T_in::tl => Some((s,-n)::nil, tl)
  | T_id s::T_num n::T_in::tl => Some((s,n)::nil, tl)
  | T_id s::T_minus::T_num n::T_semicolumn::tl =>
    match parse_variables tl with
      Some(l,r) => Some((s,-n)::l,r)
    | None => None
    end
  | T_id s::T_num n::T_semicolumn::tl =>
    match parse_variables tl with
      Some(l,r) => Some((s,n)::l,r)
    | None => None
    end
  | _ => None
  
  end.

Definition parse_right_expr (f:list token -> nat -> option(aexpr*list token)):
  list token -> nat -> aexpr -> option (aexpr*list token) :=
  fix p_r_e (l:list token) (n:nat)(acc: aexpr) {struct n} :
      option (aexpr*list token) :=
    match n with
      0%nat => None
    | S p =>
      match l with
        T_plus::T_minus::T_num n::tl => p_r_e tl p (aplus acc (anum (-n)))
      | T_plus::t_atom::tl =>
        match t_atom with
          T_num n => p_r_e tl p (aplus acc (anum n))
        | T_id s => p_r_e tl p (aplus acc (avar s))
        | T_open => match f tl p with
                      Some(e,T_close::tl') => p_r_e tl' p (aplus acc e)
                    | _ => None
                    end
        | _ => None
        end
      | T_mult::T_minus::T_num n::tl => p_r_e tl p (amult acc (anum (-n)))
      | T_mult::t_atom::tl =>
        match t_atom with
          T_num n => p_r_e tl p (amult acc (anum n))
        | T_id s => p_r_e tl p (amult acc (avar s))
        | T_open => match f tl p with
                      Some(e,T_close::tl') => p_r_e tl' p (amult acc e)
                    | _ => None
                    end
        | _ => None
        end
      | _ => Some(acc, l)
      end
    end.

Fixpoint parse_expr (l:list token)(n:nat){struct n}
   : option(aexpr*list token) :=
  match n with
    0%nat => None
  | S p =>
    match l with
      T_num n::tl => parse_right_expr parse_expr tl p (anum n)
    | T_minus::T_num n::tl => parse_right_expr parse_expr tl p (anum (-n))
    | T_id s::tl => parse_right_expr parse_expr tl p (avar s)
    | T_open::tl => match parse_expr tl p with
                      Some(e,T_close::tl') =>
                      parse_right_expr parse_expr tl' p e
                    | _ => None
                    end
    | _ => None
    end
  end.

Fixpoint parse_expr_list(l:list token)(n:nat) {struct n}:
  (list aexpr)*list token :=
  match n with 
    0%nat => (nil,l)
  | S p =>
    match parse_expr l n with
      Some(e,tl) =>
      match tl with
        T_comma::tl =>
        let (l,tl) := parse_expr_list tl p in (e::l,tl)
      | _ => (e::nil, tl)
      end
    | None => (nil, l)
    end
  end.

Definition parse_bexpr (l:list token) (n:nat) : option(bexpr * list token) :=
  match parse_expr l n with
    Some(e1, T_lt::l') =>
    match parse_expr l' n with
      Some (e2, l'') => Some (blt e1 e2, l'')
    | _ => None
    end
  | _ => None
  end.

Definition parse_elem_instr 
   (parse_instr:list token -> nat -> option(instr*list token))
   (l:list token)(n:nat) :option(instr*list token) :=
 match l with
 | T_id var::T_assign::tl =>
   match parse_expr tl n with
     Some(e,tl) => Some(assign var e, tl)
   | None => None
   end
 | T_while::tl =>
   match parse_bexpr tl n with
     Some(b,T_do::tl) =>
     match parse_instr tl n with
       Some(i,T_done::tl) =>
       Some(while b i, tl)
     | _ => None
     end
   | _ => None
   end
 | T_open_B::tl =>
   match parse_instr tl n with
     Some(i,T_close_B::tl) => Some(i,tl)
   | _ => None
   end
 | _ => None
 end.

Fixpoint parse_elem_assert (l:list token) (n:nat) {struct n}:
  option(assert*list token) :=
match parse_bexpr l n with
  Some(e, tl) => Some(a_b e, tl)
| None =>
  match n with
    0%nat => None
  | S p => 
    match l with
    | T_id s::tl => 
      if string_dec s "true" then Some(a_true, tl)
      else if string_dec s "false" then Some(a_false, tl)
      else match tl with
             T_open::tl =>
             let (l, tl) := parse_expr_list tl n in
             match tl with
               T_close::tl => Some(pred s l, tl)
             | _ => None
             end
           | _ => None
           end
    | T_bang::l => 
      match parse_elem_assert l p with
        Some(a, l) => Some(a_not a, l)
      | _ => None
      end
    | _ => None
    end
  end
end.

Fixpoint parse_assert (l:list token)(n:nat) {struct n}:
  option(assert*list token) :=
match n with 
  0%nat => None
| S p =>
  match parse_elem_assert l p with
    Some(a1, T_conj::tl) => 
    match parse_assert tl p with
      Some(a2, tl) => Some(a_conj a1 a2, tl)
    | None => None
    end
  | Some(a, tl) => Some(a, tl)
  | None => None
  end
end.

Definition parse_elem_a_instr 
   (parse_a_instr:list token -> nat -> option(a_instr*list token))
   (l:list token)(n:nat) :option(a_instr*list token) :=
 match l with
 | T_id var::T_assign::tl =>
   match parse_expr tl n with
     Some(e,tl) => Some(a_assign var e, tl)
   | None => None
   end
 | T_while::tl =>
   match parse_bexpr tl n with
     Some(b,T_do::T_open_S::tl) =>
     match parse_assert tl n with
       Some (a, T_close_S::tl) =>
       match parse_a_instr tl n with
         Some(i,T_done::tl) =>
         Some(a_while b a i, tl)
       | _ => None
       end
     | _ => None
     end
   | _ => None
   end
 | T_open_S::tl =>
   match parse_assert tl n with
     Some(a, T_close_S::tl) =>
     match parse_a_instr tl n with
       Some(i,tl) => Some(prec a i, tl)
     | _ => None
     end
   | _ => None
   end
 | T_open_B::tl =>
   match parse_a_instr tl n with
     Some(i,T_close_B::tl) => Some(i,tl)
   | _ => None
   end
 | _ => None
 end.

Fixpoint parse_instr (l:list token) (n:nat) {struct n} : 
  option(instr*list token) :=
  match n with
    0%nat => None
  | S p => 
    match parse_elem_instr parse_instr l p with
      Some(i,tl) as it =>
      match tl with
        T_semicolumn::tl =>
        match parse_instr tl p with
          Some(i', tl) =>
          Some(sequence i i', tl)
        | None => None
        end
      | _ => it
      end
    | _ => None
    end
  end.

Fixpoint parse_a_instr (l:list token) (n:nat) {struct n} :
  option(a_instr*list token) :=
  match n with
  | S p =>
    match parse_elem_a_instr parse_a_instr l p with
      Some(i, tl) as it =>
      match tl with 
        T_semicolumn::tl =>
        match parse_a_instr tl p with
          Some(i', tl) =>
          Some(a_seq i i', tl)
        | None => None
        end
      | _ => it
      end 
    | _ => None
    end
  | _ => None
  end.

Definition parse_program (l:list token) :
   option((list(string*Z))*instr) :=
   match l with
     T_variables::tl =>
     match parse_variables tl with
       Some(l,r) => 
        match parse_instr r (List.length r) with
          Some (i, T_end::nil) => Some(l,i)
        | _ => None
        end
     | None => None
     end
   | _ => None
   end.

Definition parse (s : string) : option((list(string*Z))*instr) :=
  parse_program (tokenize s Start "").

Definition parse_i (s : string) : instr :=
  let ts := tokenize s Start "" in
  match parse_instr ts (List.length ts) with
    Some (a, nil) => a | _ => assign "parse failed" (anum 0) end.

Definition parse_ai' (s :string) : a_instr :=
  let ts := tokenize s Start "" in
  match parse_a_instr ts (List.length ts) with
    Some (a, nil) => a | _ => a_assign "parse failed" (anum 0)
  end.

Definition parse_e (s : string) : option aexpr :=
  let ts := tokenize s Start "" in
  match parse_expr ts (S (List.length ts)) with
    Some (a, nil) => Some a | _ => None end.

Definition eol := "
".

Fixpoint string_of_env (l:list(string*Z)) : string :=
  match l with
  | (s,n)::nil => s ++ " " ++ string_of_z n
  | (s,n)::tl =>
    s ++ " " ++ string_of_z n ++ ";" ++ eol ++ string_of_env tl
  | nil => ""
  end.

Fixpoint string_of_aexpr (e:aexpr) : string :=
  match e with
    anum x => string_of_z x
  | avar a => a
  | aplus e1 ((aplus _ _) as e2) => string_of_aexpr e1 ++ "+(" ++
                                     string_of_aexpr e2 ++ ")"
  | amult e1 ((aplus _ _) as e2) => string_of_aexpr e1 ++ "*(" ++
                                     string_of_aexpr e2 ++ ")"
  | aplus e1 e2 => string_of_aexpr e1 ++ " + " ++ string_of_aexpr e2
  | amult e1 e2 => string_of_aexpr e1 ++ " * " ++ string_of_aexpr e2
  end.

Definition string_of_bexpr (e:bexpr) :string :=
  match e with
    blt e1 e2 => string_of_aexpr e1 ++ " < " ++ string_of_aexpr e2
  end.

Fixpoint string_of_instr (i:instr) :string :=
  match i with
  | sequence ((sequence _ _) as i1) i2 => 
     "{" ++ string_of_instr i1 ++ "};" ++ eol ++ string_of_instr i2
  | sequence i1 i2 => 
    string_of_instr i1 ++ ";" ++ eol ++ string_of_instr i2
  | assign x e => x ++ ":=" ++ string_of_aexpr e
  | while b i => "while " ++ string_of_bexpr b ++ " do" ++ eol ++
                  string_of_instr i ++ eol ++ "done" ++ eol
  end.


Ltac parse_it :=
  match goal with
  | |- context[parse_instr ?x] =>
    let v := eval vm_compute in (parse_instr x) in
      change (parse_instr x) with v; parse_it
  | _ => idtac
  end.
